/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/model/tablecollection', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/object/timermixin',
    'io.ox/office/baseframework/model/modelobject',
    'io.ox/office/editframework/model/attributedmodel',
    'io.ox/office/spreadsheet/utils/operations',
    'io.ox/office/spreadsheet/utils/sheetutils'
], function (Utils, TimerMixin, ModelObject, AttributedModel, Operations, SheetUtils) {

    'use strict';

    var // convenience shortcuts
        Interval = SheetUtils.Interval,
        Range = SheetUtils.Range,
        IntervalArray = SheetUtils.IntervalArray,
        RangeArray = SheetUtils.RangeArray;

    // class TableColumnModel =================================================

    /**
     * Stores filter and sorting settings for a single column in a table range.
     *
     * @constructor
     *
     * @extends AttributedModel
     *
     * @param {SheetModel} sheetModel
     *  The sheet model instance containing this table column.
     *
     * @param {SheetModel} tableModel
     *  The table model instance containing this table column.
     *
     * @param {Object} [initAttrs]
     *  The initial table attribute set (style sheet reference, and explicit
     *  table attributes).
     */
    var TableColumnModel = AttributedModel.extend({ constructor: function (sheetModel, tableModel, initAttrs) {

        // base constructor ---------------------------------------------------

        AttributedModel.call(this, sheetModel.getDocModel(), initAttrs, { families: 'filter sort', autoClear: true });

        // protected methods --------------------------------------------------

        /**
         * Creates and returns a cloned instance of this table model for the
         * specified sheet.
         *
         * @internal
         *  Used by the class TableCollection during clone construction. DO NOT
         *  CALL from external code!
         *
         * @param {SheetModel} targetSheetModel
         *  The model instance of the new cloned sheet that will own the clone
         *  returned by this method.
         *
         * @param {TableModel} targetTableModel
         *  The model instance of the new cloned table that will own the clone
         *  returned by this method.
         *
         * @returns {TableColumnModel}
         *  A clone of this table column model, initialized for ownership by
         *  the passed sheet and table models.
         */
        this.clone = function (targetSheetModel, targetTableModel) {
            // construct a new model instance
            return new TableColumnModel(targetSheetModel, targetTableModel, this.getExplicitAttributes(true));
        };

        // public methods -----------------------------------------------------

        /**
         * Returns whether this column model contains active filter rules.
         *
         * @returns {Boolean}
         *  Whether this column model contains active filter rules.
         */
        this.isFiltered = function () {
            return this.getMergedAttributes().filter.type !== 'none';
        };

        /**
         * Returns whether this column model contains active sort options.
         *
         * @returns {Boolean}
         *  Whether this column model contains active sort options.
         */
        this.isSorted = function () {
            return false; // TODO
        };

        /**
         * Returns whether this column model contains any filter rules, or sort
         * options, that may require to refresh the table range containing this
         * column.
         *
         * @returns {Boolean}
         *  Whether this column model contains any filter/sorting settings.
         */
        this.isRefreshable = function () {
            return this.isFiltered() || this.isSorted();
        };

        // initialization -----------------------------------------------------

        // destroy all class members
        this.registerDestructor(function () {
            sheetModel = tableModel = null;
        });

    } }); // class TableColumnModel

    // class TableModel =======================================================

    /**
     * Stores settings for a single table range in a specific sheet.
     *
     * @constructor
     *
     * @extends AttributedModel
     * @extends TimerMixin
     *
     * @param {SheetModel} sheetModel
     *  The sheet model instance containing this table.
     *
     * @param {String} tableName
     *  The name of this table.
     *
     * @param {Range} tableRange
     *  The address of the cell range covered by this table.
     *
     * @param {Object} [initAttrs]
     *  The initial table attribute set (style sheet reference, and explicit
     *  table attributes).
     */
    var TableModel = AttributedModel.extend({ constructor: function (sheetModel, tableName, tableRange, initAttrs) {

        var // self reference
            self = this,

            // the spreadsheet document model
            docModel = sheetModel.getDocModel(),

            // all existing column models (with active filter or sorting settings), mapped by column index (for fast direct access)
            columnModels = {},

            // default column model for all unmodified columns
            defColumnModel = new TableColumnModel(sheetModel, this);

        // base constructors --------------------------------------------------

        AttributedModel.call(this, docModel, initAttrs, { styleFamily: 'table', autoClear: true });
        TimerMixin.call(this);

        // private methods ----------------------------------------------------

        /**
         * Inserts the passed table column model into this table, and registers
         * a listener for attribute change events that will be forwarded to the
         * listeners of this table.
         */
        function insertColumnModel(tableCol, columnModel) {

            var // the map entry, containing the current table column index as property (may change during runtime,
                // e.g. after inserting/deleting sheet columns, needed for fast access when forwarding events below)
                entry = { tableCol: tableCol, columnModel: columnModel };

            // store the column model as map entry (needed for fast direct access by column index)
            columnModels[tableCol] = entry;

            // listen to attribute change events, forward to own listeners
            self.listenTo(columnModel, 'change:attributes', function (event, newAttrs, oldAttrs) {

                // forward to listeners of this table model (pass current column index, not the
                // original column index passed to the method insertColumnModel(), see comment above)
                self.trigger('change:column', entry.tableCol, newAttrs, oldAttrs);

                // model can be deleted, if it does not contain any active settings
                if (!columnModel.hasExplicitAttributes()) {
                    columnModel.destroy();
                    delete columnModels[entry.tableCol];
                }
            });
        }

        /**
         * Returns the specified table column model, if it exists.
         */
        function getColumnModel(tableCol) {
            var entry = columnModels[tableCol];
            return entry ? entry.columnModel : null;
        }

        /**
         * Creates a missing table column model, and returns the model
         * instance. The passed column index will not be checked for validity.
         */
        function getOrCreateColumnModel(tableCol) {

            var // find an existing column model
                columnModel = getColumnModel(tableCol);

            // create missing column model
            if (!columnModel) {
                columnModel = new TableColumnModel(sheetModel, self);
                insertColumnModel(tableCol, columnModel);
            }

            return columnModel;
        }

        /**
         * Updates internal settings after the attributes of this table have
         * been changed.
         */
        function changeAttributesHandler() {

            // silently remove filter rules from all filtered column models, when
            // table is not active anymore (sorting settings remain valid though)
            if (!self.isFilterActive()) {
                _.each(columnModels, function (entry, key) {
                    // remove filter settings from the column
                    entry.columnModel.setAttributes({ filter: { type: 'none' } }, { notify: 'never' });
                    // keep the column model in the array, if any attributes are left
                    if (!entry.columnModel.hasExplicitAttributes()) {
                        entry.columnModel.destroy();
                        delete columnModels[key];
                    }
                });
            }
        }

        /**
         * Creates a filter matcher for the passed entries of a discrete filter
         * rule.
         *
         * @param {Array<String>} entries
         *  The matching entries of a discrete filter rule.
         *
         * @returns {Function}
         *  A predicate function that expects the typed cell result value, and
         *  the formatted display string as parameters, and returns a Boolean
         *  whether the passed cell value matches any of the entries in the
         *  string array.
         */
        function createDiscreteFilterMatcher(entries) {

            // no matching entries: all cell values will be filtered
            if (entries.length === 0) {
                return _.constant(false);
            }

            // performance: create a map with all matching display strings
            // as keys, instead of searching in the entry array
            var entryMap = {};
            if (docModel.getApp().isOOXML()) {
                entryMap = {};
                _.each(entries, function (entry) { entryMap[entry.toLowerCase()] = 1; });
            } else {
                _.each(entries, function (entry) { entryMap[getODFEntryKey(entry)] = 1; });
            }

            if (docModel.getApp().isODF()) {
                // create a matcher predicate function that checks the lower-case display string of a cell
                return function discreteFilterMatcher(result) {
                    return getODFEntryKey(result) in entryMap;
                };
            } else {
                // create a matcher predicate function that checks the lower-case display string of a cell
                return function discreteFilterMatcher(result, display) {
                    // bug 36264: trim space characters but no other whitespace
                    return display.replace(/^ +| +$/g, '').toLowerCase() in entryMap;
                };
            }

        }

        /**
         * Return the entrie key for odf values, it's a combination of the result value
         * and the type of the result object e.g. String "foo" = "foo_s".
         * @param {Object} result value to get the key
         * @returns {String} the key
         */
        function getODFEntryKey(result) {
            var key = result + '_';
            if (_.isNull(result)) {
                key += 'b';
            } else if (_.isNumber(result)) {
                key += 'n';
            } else if (_.isBoolean(result)) {
                key += 'boo';
            } else if (_.isString(result)) {
                key = key.toLowerCase();
                key += 's';
            } else {
                key += 'e';
            }

            return key;
        }

        // protected methods --------------------------------------------------

        /**
         * Creates and returns a cloned instance of this table model for the
         * specified sheet.
         *
         * @internal
         *  Used by the class TableCollection during clone construction. DO NOT
         *  CALL from external code!
         *
         * @param {SheetModel} targetModel
         *  The model instance of the new cloned sheet that will own the clone
         *  returned by this method.
         *
         * @returns {TableModel}
         *  A clone of this table model, initialized for ownership by the
         *  passed sheet model.
         */
        this.clone = function (targetModel) {
            // construct a new model instance, pass all own column models as hidden parameter
            return new TableModel(targetModel, tableName, tableRange, this.getExplicitAttributes(true), columnModels);
        };

        /**
         * Callback handler for the document operation 'changeTableColumn'.
         * Changes the attributes of a column in this table model.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'changeTableColumn' document operation.
         */
        this.applyChangeOperation = function (context) {

            // check that the column index exists in the table
            var tableCol = context.getInt('col');
            context.ensure((tableCol >= 0) && (tableCol < tableRange.cols()), 'invalid table column');

            // change the column attributes
            var attributes = context.getObj('attrs');
            getOrCreateColumnModel(tableCol).setAttributes(attributes);
        };

        // public methods -----------------------------------------------------

        /**
         * Returns the name of this table.
         *
         * @returns {String}
         *  The name of this table.
         */
        this.getName = function () {
            return tableName;
        };

        /**
         * Returns whether this table id the special auto-filter range of the
         * sheet.
         *
         * @returns {Boolean}
         *  Whether this table id the special auto-filter range of the sheet.
         */
        this.isAutoFilter = function () {
            return tableName.length === 0;
        };

        /**
         * Returns the address of the cell range covered by this table.
         *
         * @returns {Range}
         *  The address of the cell range covered by this table.
         */
        this.getRange = function () {
            return tableRange.clone();
        };

        /**
         * Returns whether the passed cell address is covered by this table.
         *
         * @param {Address} address
         *  The address of the cell to be checked.
         *
         * @returns {Boolean}
         *  Whether the specified cell is covered by this table.
         */
        this.containsCell = function (address) {
            return tableRange.containsAddress(address);
        };

        /**
         * Returns whether the passed cell range address overlaps with the cell
         * range of this table.
         *
         * @param {Range} range
         *  The address of the cell range to be checked.
         *
         * @returns {Boolean}
         *  Whether the passed cell range address overlaps with the cell range
         *  of this table.
         */
        this.overlapsRange = function (range) {
            return tableRange.overlaps(range);
        };

        /**
         * Returns whether this table range contains a header row. Cells in the
         * header row (the top-most row) will neither be filtered nor sorted.
         *
         * @returns {Boolean}
         *  Whether this table range contains a header row.
         */
        this.hasHeaderRow = function () {
            return this.getMergedAttributes().table.headerRow;
        };

        /**
         * Returns whether this table range contains a footer row. Cells in the
         * footer row (the bottom-most row) will neither be filtered nor
         * sorted.
         *
         * @returns {Boolean}
         *  Whether this table range contains a footer row.
         */
        this.hasFooterRow = function () {
            return this.getMergedAttributes().table.footerRow;
        };

        /**
         * Returns whether filtering in this table range is activated. This
         * does not necessarily mean that a row has actually been hidden by the
         * filter rules.
         *
         * @returns {Boolean}
         *  Whether filtering in this table range is activated.
         */
        this.isFilterActive = function () {
            return this.getMergedAttributes().table.filtered;
        };

        /**
         * Returns the address of the data range in this table (the cell range
         * without the header and footer rows). For the special auto-filter
         * table, the data range will be expanded to the bottom to include all
         * existing content cells following the real data range.
         *
         * @returns {Range|Null}
         *  The address of the data range of this table; or null, if no data
         *  rows are available (e.g., if the table consists of a header row
         *  only).
         */
        this.getDataRange = function () {

            var // whether the table contains header and/or footer rows
                hasHeader = this.hasHeaderRow(),
                hasFooter = this.hasFooterRow(),
                // the resulting data range
                dataRange = tableRange.clone();

            // expand data range of auto-filter to the bottom (before removing
            // the header row from the range, to not invalidate the range address)
            if (!hasFooter && this.isAutoFilter()) {
                dataRange = sheetModel.getCellCollection().findContentRange(dataRange, { directions: 'down', skipHidden: true });

                // restrict expansion to the next table range below the auto-filter
                if (tableRange.end[1] < dataRange.end[1]) {
                    var boundRange = Range.create(tableRange.start[0], tableRange.end[1] + 1, tableRange.end[0], docModel.getMaxRow());
                    sheetModel.getTableCollection().findTables(boundRange).forEach(function (otherTable) {
                        var otherTableRange = otherTable.getRange();
                        tableRange.end[1] = Math.min(tableRange.end[1], otherTableRange.start[1] - 1);
                    });
                }
            }

            // exclude header/footer rows from returned range
            if (hasHeader) { dataRange.start[1] += 1; }
            if (hasFooter) { dataRange.end[1] -= 1; }

            // check existence of data rows
            return (dataRange.start[1] <= dataRange.end[1]) ? dataRange : null;
        };

        /**
         * Returns the relative intervals of all visible rows in the data range
         * of this table.
         *
         * @returns {IntervalArray}
         *  The visible row intervals in the data range of this table; or an
         *  empty array, if the table does not contain any data rows. The
         *  properties 'first' and 'last' of each index interval are relative
         *  to the data range of the table (the index 0 refers to the first
         *  data row of this table!).
         */
        this.getVisibleDataRows = function () {

            var // the address of the data range
                dataRange = this.getDataRange();

            // return empty interval array, if data range is missing
            if (!dataRange) { return new IntervalArray(); }

            var // the absolute row intervals
                rowIntervals = sheetModel.getRowCollection().getVisibleIntervals(dataRange.rowInterval());

            // transform to relative row intervals
            return rowIntervals.move(-dataRange.start[1]);
        };

        /**
         * Changes the address of the cell range covered by this table.
         *
         * @param {Range} range
         *  The address of the new cell range.
         *
         * @returns {TableModel}
         *  A reference to this instance.
         */
        this.setRange = function (range) {

            if (tableRange.equals(range)) { return this; }

            // store the new range (use a copy to prevent indirect modification)
            tableRange = range.clone();

            // delete table columns outside the new range
            var colCount = tableRange.cols();
            _.each(columnModels, function (entry, key) {
                if (entry.tableCol >= colCount) {
                    entry.columnModel.destroy();
                    delete columnModels[key];
                }
            });

            // notify listeners
            this.trigger('change:range', range);
            return this;
        };

        /**
         * Transforms the cell range address of this table, after columns or
         * rows have been inserted into or removed from the sheet.
         *
         * @returns {Boolean}
         *  Whether the transformed range is still valid (not completely
         *  deleted). If false is returned, this table model must be deleted
         *  from the table collection.
         */
        this.transformRange = function (interval, insert, columns) {

            var // first column index of the old table range
                startCol = tableRange.start[0];

            // preprocessing when deleting rows in the table range
            if (!insert && !columns) {

                // bug 36226: delete tables whose header row has been deleted
                // bug 39869: deleting header rows prevented by UI except for auto-filter
                if (this.hasHeaderRow() && interval.containsIndex(tableRange.start[1])) {
                    return false;
                }

                // update existence of footer row
                if (this.hasFooterRow() && interval.containsIndex(tableRange.end[1])) {
                    this.setAttributes({ table: { footerRow: false } }, { notify: 'never' });
                }
            }

            // transform table range
            tableRange = docModel.transformRange(tableRange, interval, insert, columns);

            // table completely deleted
            if (!tableRange) { return false; }

            // delete or remap all affected column models
            if (columns) {

                var // create a new model map to prevent interfering with existing map entries
                    newColumnModels = {};

                _.each(columnModels, function (entry) {

                    var // transform the absolute column index according to the operation
                        absCol = docModel.transformIndex(startCol + entry.tableCol, interval, insert, true);

                    // delete column model of deleted columns, remap remaining column models
                    if (_.isNumber(absCol)) {
                        entry.tableCol = absCol - tableRange.start[0];
                        newColumnModels[entry.tableCol] = entry;
                    } else {
                        entry.columnModel.destroy();
                    }
                });

                columnModels = newColumnModels;
            }

            return true;
        };

        /**
         * Returns the merged column attributes of the specified table column.
         *
         * @param {Number} tableCol
         *  The zero-based index of the table column, relative to the cell
         *  range covered by this table.
         *
         * @returns {Object|Null}
         *  The merged attribute set of the specified table column; or null, if
         *  the passed table column index is invalid.
         */
        this.getColumnAttributes = function (tableCol) {

            var // the column model (fall-back to default model for unmodified columns)
                columnModel = getColumnModel(tableCol) || defColumnModel;

            return columnModel.getMergedAttributes();
        };

        /**
         * Returns the address of the data range of a single column in this
         * table, without the header and footer cell.
         *
         * @param {Number} tableCol
         *  The zero-based index of the column, relative to the cell range
         *  covered by this table.
         *
         * @returns {Range|Null}
         *  The address of the data range in the specified table column; or
         *  null, if no data rows are available (e.g., if the table consists of
         *  a header row only), or if the passed column index is invalid.
         */
        this.getColumnDataRange = function (tableCol) {

            var // the data range of the entire table
                dataRange = this.getDataRange(),
                // the resulting absolute column index
                absCol = tableRange.start[0] + tableCol;

            // check validity of passed column index, and existence of data rows
            if (!dataRange || !tableRange.containsCol(absCol)) {
                return null;
            }

            // adjust column indexes in the result range
            return dataRange.setBoth(absCol, true);
        };

        /**
         * Returns the cell contents of the data range of a single column in
         * this table.
         *
         * @param {Number} tableCol
         *  The zero-based index of the column, relative to the cell range
         *  covered by this table.
         *
         * @param {IntervalArray|Interval} [rowIntervals]
         *  If specified, an array of row intervals, or a single row interval,
         *  in the data range of this table to query the cell data for. The
         *  properties 'first' and 'last' of each interval object are relative
         *  to the data range of the table (the index 0 refers to the first
         *  data row of this table!). If omitted, returns the cell contents of
         *  all data cells in the column.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved with an
         *  compressed array of cell contents in the specified table column.
         *  Each array element is an object with the following properties:
         *  - {String|Null} cellData.display
         *      The display string of the cell result (will be the empty string
         *      for blank cells). The value null represents a cell result that
         *      cannot be formatted to a valid display string with the current
         *      number format of the cell.
         *  - {Number|String|Boolean|ErrorCode|Null} cellData.result
         *      The typed result value (will be null for blank cells).
         *  - {Number} cellData.count
         *      The number of consecutive cells with equal contents represented
         *      by this array element.
         *  See method CellCollection.queryCellContents() for more details.
         */
        this.queryColumnData = function (tableCol, rowIntervals) {

            var // the address of the data range in the specified column
                dataRange = this.getColumnDataRange(tableCol);

            // reject the request, if no data range is available (e.g., header/footer rows only)
            if (!dataRange) { return $.when([]); }

            var // the row interval of the entire data range
                boundInterval = dataRange.rowInterval(),
                // the cell ranges to be passed to the query
                dataRanges = null;

            // transform the passed relative intervals to absolute row intervals
            if (rowIntervals) {
                rowIntervals = rowIntervals.clone(true).move(boundInterval.first).intersect(boundInterval);
            } else {
                rowIntervals = new IntervalArray(boundInterval);
            }

            // nothing to do, if the passed relative intervals are outside the data range
            if (rowIntervals.empty()) { return $.when([]); }

            // build the data range addresses to be queried for
            dataRanges = RangeArray.createFromRowIntervals(rowIntervals, dataRange.start[0]);

            // resolve the array of result arrays returned by the cell collection to a single result array
            // Bug 40436: restrict the query to 3 server requests, as workaround for broken ODS documents
            // reporting the entire sheet as used area
            return sheetModel.getCellCollection()
                .queryCellContents([dataRanges], { hidden: true, compressed: true, maxRequests: 3 })
                .then(function (resultsArray) { return resultsArray[0]; });
        };

        /**
         * Returns whether the specified column contains active filter rules.
         *
         * @param {Number} tableCol
         *  The zero-based index of the column, relative to the cell range
         *  covered by this table.
         *
         * @returns {Boolean}
         *  Whether the specified column contains active filter rules.
         */
        this.isColumnFiltered = function (tableCol) {
            var columnModel = getColumnModel(tableCol);
            return _.isObject(columnModel) && columnModel.isFiltered();
        };

        /**
         * Returns whether the specified column contains active sort options.
         *
         * @param {Number} tableCol
         *  The zero-based index of the column, relative to the cell range
         *  covered by this table.
         *
         * @returns {Boolean}
         *  Whether the specified column contains active sort options.
         */
        this.isColumnSorted = function (tableCol) {
            var columnModel = getColumnModel(tableCol);
            return _.isObject(columnModel) && columnModel.isSorted();
        };

        /**
         * Returns whether this table contains any columns with active settings
         * (filter rules or sort options) that can be used to refresh the table
         * range.
         *
         * @returns {Boolean}
         *  Whether this table contains any columns with active settings.
         */
        this.isRefreshable = function () {
            return this.isFilterActive() && _.any(columnModels, function (entry) {
                return entry.columnModel.isRefreshable();
            });
        };

        /**
         * Creates a matcher predicate function according to the filter
         * attributes of the specified table column.
         *
         * @param {Number} tableCol
         *  The zero-based index of the column, relative to the cell range
         *  covered by this table.
         *
         * @returns {Function|Null}
         *  A predicate function that takes a cell descriptor object with the
         *  properties 'result' (typed cell result value) and 'display' (the
         *  formatted display string), and returns a Boolean whether the passed
         *  cell matches the filter rules of the specified column. This method
         *  will return null instead of a matcher function, if the specified
         *  table column does not contain any supported filter rules.
         */
        this.createFilterMatcher = function (tableCol) {

            var // the current filter attributes of the passed column
                filterAttrs = this.getColumnAttributes(tableCol).filter;

            // create a matcher function for discrete filter rule
            if (filterAttrs.type === 'discrete') {
                return createDiscreteFilterMatcher(filterAttrs.entries);
            }

            // unsupported filter type, or no existing filter rule
            return null;
        };

        /**
         * Builds and returns a list of row intervals containing all visible
         * rows not filtered by the current filter settings. For performance
         * reasons (to prevent blocked browser UI and its respective error
         * messages), the implementation of this method may run asynchronously.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Number} [options.excludeTableCol]
         *      The zero-based index of a table column that will be excluded
         *      from filtering (its filter settings will be ignored).
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved with the
         *  visible data row intervals (instance of the class IntervalArray).
         *  The properties 'first' and 'last' of each row interval are relative
         *  to the data range of the table (the index 0 always refers to the
         *  first data row of this table!).
         */
        this.queryFilterResult = function (options) {

            var // start with the entire data range
                dataRange = this.getDataRange();

            // no data rows available
            if (!dataRange) { return $.when([]); }

            var // indexes of all filtered columns
                filterCols = [],
                // table column to be excluded from filtering
                excludeTableCol = Utils.getIntegerOption(options, 'excludeTableCol', -1),
                // the entire row interval of the data range (absolute row indexes)
                dataRowInterval = dataRange.rowInterval(),
                // the remaining visible data intervals for the next column (relative to the data range)
                visibleIntervals = new IntervalArray(new Interval(0, dataRowInterval.size() - 1)),
                // the resulting promise of this method
                promise = null;

            // collect the column indexes of all filtered columns
            if (this.isFilterActive()) {
                _.each(columnModels, function (entry) {
                    if (entry.columnModel.isFiltered() && (entry.tableCol !== excludeTableCol)) {
                        filterCols.push(entry.tableCol);
                    }
                });
            }

            // process the table columns in a background loop to prevent unresponsive browser
            promise = this.iterateArraySliced(filterCols, function (tableCol) {

                // early exit, if no visible data rows are left for filtering
                if (visibleIntervals.empty()) { return Utils.BREAK; }

                var // the filter matcher for the current column
                    matcher = this.createFilterMatcher(tableCol);

                // ignore columns with unknown/unsupported filter settings
                if (!matcher) { return; }

                // query the cell contents of all remaining visible rows
                return this.queryColumnData(tableCol, visibleIntervals).then(function (cellArray) {

                    var // the resulting row intervals which remain visible
                        newIntervals = new IntervalArray(),
                        // index of the current data row
                        dataRow = 0,
                        // index of the current row interval, and offset inside the interval
                        intervalIndex = 0;

                    // pushes a new row interval to the 'newRowIntervals' array (tries to expand last interval)
                    function pushVisibleInterval(cellCount) {

                        var // end position of the new interval (start position is 'dataRow')
                            lastDataRow = dataRow + cellCount - 1,
                            // last pushed interval for expansion
                            lastInterval = newIntervals.last();

                        // try to expand the last interval
                        if (lastInterval && (lastInterval.last + 1 === dataRow)) {
                            lastInterval.last = lastDataRow;
                        } else {
                            newIntervals.push(new Interval(dataRow, lastDataRow));
                        }
                    }

                    // simultaneously iterate through the row intervals and compressed result array
                    cellArray.forEach(function (cellData) {

                        var // the number of remaining cell to process for the current cell data object
                            remainingCells = cellData.count,
                            // whether the current cell data results in visible rows (the filter matches)
                            visible = _.isString(cellData.display) && matcher(cellData.result, cellData.display),
                            // the current row interval
                            visibleInterval = null,
                            // number of cells that can be processed for the current row interval
                            cellCount = 0;

                        // process as many remaining data rows as covered by the current cell data entry
                        while (remainingCells > 0) {

                            // the current row interval
                            visibleInterval = visibleIntervals[intervalIndex];
                            // update 'dataRow' if a new row interval has been started
                            dataRow = Math.max(dataRow, visibleInterval.first);
                            // number of cells that can be processed for the current row interval
                            cellCount = Math.min(remainingCells, visibleInterval.last - dataRow + 1);

                            // if the cell data matches the filter rule, add the row interval to the result
                            if (visible) { pushVisibleInterval(cellCount); }

                            // decrease number of remaining cells for the current cell data object
                            remainingCells -= cellCount;
                            // update data row index
                            dataRow += cellCount;
                            // if end of current row interval has been reached, go to next row interval
                            if (visibleInterval.last < dataRow) { intervalIndex += 1; }
                        }
                    });

                    // use the resulting filtered row intervals for the next filtered column
                    visibleIntervals = newIntervals;
                });
            });

            // return a promise that resolves with the row intervals
            return promise.then(function () { return visibleIntervals; });
        };

        // initialization -----------------------------------------------------

        // store a clone of the range to prevent external modifications
        tableRange = tableRange.clone();

        // update column models after changing table attributes
        this.on('change:attributes', changeAttributesHandler);

        // clone column models passed as hidden argument by the clone() method
        if (_.isObject(arguments[TableModel.length])) {
            _.each(arguments[TableModel.length], function (entry) {
                insertColumnModel(entry.tableCol, entry.columnModel.clone(sheetModel, this));
            }, this);
        }

        // destroy all class members
        this.registerDestructor(function () {
            _.each(columnModels, function (entry) { entry.columnModel.destroy(); });
            defColumnModel.destroy();
            docModel = sheetModel = columnModels = defColumnModel = null;
        });

    } }); // class TableModel

    // class TableCollection ==================================================

    /**
     * Stores settings for table ranges in a specific sheet. Table ranges
     * contain filter settings, sorting settings, and specific cell formatting.
     *
     * Triggers the following events:
     * - 'insert:table'
     *      After a new table range has been inserted into this collection.
     *      Event handlers receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {String} key
     *          The unique key of the new table range.
     * - 'change:table'
     *      After an existing table range has been changed. Event handlers
     *      receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {String} key
     *          The unique key of the changed table range.
     *      (3) {String} type
     *          The type of the original event forwarded from the table model.
     *          Can be 'change:attributes', 'change:range', or 'change:column'.
     *      (4) {Any} [...]
     *          Additionally parameters of the original table event.
     * - 'delete:table'
     *      After an existing table range has been deleted from this
     *      collection. Event handlers receive the following parameters:
     *      (1) {jQuery.Event} event
     *          The jQuery event object.
     *      (2) {String} key
     *          The unique key of the deleted table range.
     *
     * @constructor
     *
     * @extends ModelObject
     *
     * @param {SheetModel} sheetModel
     *  The sheet model instance containing this collection.
     */
    function TableCollection(sheetModel) {

        var // self reference
            self = this,

            // the document mode
            docModel = sheetModel.getDocModel(),

            // all table models, mapped by upper-case table name
            tableModels = {};

        // base constructor ---------------------------------------------------

        ModelObject.call(this, sheetModel.getDocModel());

        // private methods ----------------------------------------------------

        /**
         * Returns a descriptor for an existing table model addressed by the
         * passed operation context.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing a document operation for tables.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.insert=false]
         *      If set to true, the table model addressed by the operation must
         *      not exist in this table collection. By default, the table model
         *      must exist in the collection.
         *
         * @returns {Object}
         *  A descriptor for the table model addressed by the passed operation,
         *  with the following properties:
         *  - {String} name
         *      The original table name as contained in the passed operation.
         *  - {String} key
         *      The (upper-case) name used as map key for the table model.
         *  - {TableModel|Null} model
         *      The table model addressed by the operation; or null, if the
         *      table does not exist (see option 'insert' above).
         *
         * @throws {OperationException}
         *  If the operation does not address an existing table model, or if it
         *  addresses an existing table model that must not exist (see option
         *  'insert' above).
         */
        function getModelData(context, options) {

            // get the table name from the operation (missing/empty name addresses auto filter)
            var tableName = context.getOptStr('table', true);

            // the upper-case name, used as map key
            var tableKey = tableName.toUpperCase();

            // check existence/absence of the table model
            var tableModel = tableModels[tableKey] || null;
            if (Utils.getBooleanOption(options, 'insert', false)) {
                context.ensure(!tableModel, 'table exists');
            } else {
                context.ensure(tableModel, 'missing table');
            }

            return { name: tableName, key: tableKey, model: tableModel };
        }

        /**
         * Returns whether the passed cell range address covers a table model
         * in this collection.
         *
         * @param {Range} range
         *  The cell range address to be checked.
         *
         * @param {TableModel} [excludeModel]
         *  A table model whose cell range will be ignored.
         *
         * @returns {Boolean}
         *  Whether the passed cell range address covers a table model (except
         *  the table model specified in the parameter 'excludeModel') in this
         *  collection.
         */
        function rangeCoversAnyTable(range, excludeModel) {
            return _.some(tableModels, function (tableModel) {
                return (tableModel !== excludeModel) && tableModel.overlapsRange(range);
            });
        }

        /**
         * Inserts the passed table model into this collection, triggers an
         * 'insert:table' event, and registers a listener for all events of the
         * table model that will be forwarded to the listeners of this
         * collection.
         */
        function insertTableModel(tableModel) {

            // insert the passed table model, notify listeners
            var tableKey = tableModel.getName().toUpperCase();
            tableModels[tableKey] = tableModel;
            self.trigger('insert:table', tableKey);

            // forward all change events of the table model
            self.listenTo(tableModel, 'triggered', function (event, type) {
                self.trigger.apply(self, ['change:table', tableModel.getName().toUpperCase(), type].concat(_.toArray(arguments).slice(2)));
            });
        }

        /**
         * Deletes an existing table model from this collection, and triggers a
         * 'delete:table' event.
         */
        function deleteTableModel(tableKey) {
            tableModels[tableKey].destroy();
            delete tableModels[tableKey];
            self.trigger('delete:table', tableKey);
        }

        /**
         * Invokes the specified method on all table models in this collection.
         *
         * @param {String} methodName
         *  The name of the method to be invoked on each table model. If the
         *  method returns the value false, the respective table model will be
         *  removed from the collection.
         *
         * @param {TableModel|Null} excludeModel
         *  A table model that will be skipped.
         *
         * @param {Any} ...
         *  Other parameters that will be passed to the specified method.
         */
        function invokeModelMethod(methodName, excludeModel) {
            var args = _.toArray(arguments).slice(invokeModelMethod.length);
            _.each(tableModels, function (tableModel, tableKey) {
                if ((tableModel !== excludeModel) && (tableModel[methodName].apply(tableModel, args) === false)) {
                    deleteTableModel(tableKey);
                }
            });
        }

        /**
         * Recalculates the position of all table ranges, after columns or rows
         * have been inserted into or deleted from the sheet.
         */
        function transformRanges(interval, insert, columns) {
            invokeModelMethod('transformRange', null, interval, insert, columns);
        }

        /**
         * Invokes the passed callback function with activated custom 'action
         * processing mode' in the document model. This method receives the
         * name of a table range, and passes the table model to the callback
         * function, otherwise behaves similarly to the method
         * SheetModel.createAndSendOperationsWithDrawings().
         *
         * @param {String} tableName
         *  The name of an existing table range. The empty string refers to the
         *  anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet. If the table range does not
         *  exist, this method will exit with the error code 'table:missing'.
         *
         * @param {Function} callback
         *  The callback function that will be invoked with activated action
         *  processing mode. Receives the following parameters:
         *  (1) {SheetOperationsGenerator} generator
         *      An empty sheet operations generator.
         *  (2) {TableModel} tableModel
         *      A reference to the table model specified with the parameter
         *      'tableName' passed to this method.
         *  May return a promise to defer the action processing mode until it
         *  has been resolved or rejected.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved or rejected according to the result
         *  of the callback function. If the table name passed to this method
         *  is not valid, the promise will be rejected with the error object
         *  {cause:'table:missing'}.
         */
        function createAndSendTableOperations(tableName, callback) {

            // the table model
            var tableModel = self.getTable(tableName);

            // check existence of the table range
            if (!tableModel) { return SheetUtils.makeRejected('table:missing'); }

            // send operations without applying them (done manually during creation)
            return sheetModel.createAndSendOperationsWithDrawings(function (generator) {
                return callback.call(this, generator, tableModel);
            });
        }

        /**
         * Generates all needed operations to refresh the passed table model.
         * Evaluates the current filter settings, and updates the visibility of
         * all rows covered by the table range. Also generates all implicit
         * operations for drawing objects.
         *
         * @param {SheetOperationsGenerator} generator
         *  The operations generator that will be filled with the operations.
         *
         * @param {TableModel} tableModel
         *  The model of the table range to be refreshed.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the table has been refreshed
         *  successfully, or that will be rejected with an object with 'cause'
         *  property set to one of the following error codes:
         *  - 'rows:overflow': The current filter settings would modify (show
         *      or hide) too many rows at once.
         */
        function generateRefreshTableOperations(generator, tableModel) {

            // start with the entire data range
            var dataRange = tableModel.getDataRange();

            // do nothing, if no data rows are available
            if (!dataRange) { return $.when(); }

            var // the entire table range
                tableRange = tableModel.getRange(),
                // the effective table range, e.g. after expanding auto-filter data range
                effectiveRange = tableRange.boundary(dataRange);

            // update the effective table range (data range may have been expanded)
            if (tableRange.differs(effectiveRange)) {
                generator.generateTableOperation(Operations.CHANGE_TABLE, tableModel.getName(), effectiveRange.toJSON());
                docModel.invokeOperationHandler(generator.getLastOperation());
            }

            // update the visibility state of all filter rows
            return tableModel.queryFilterResult().then(function (rowIntervals) {

                var // the entire row interval of the data range
                    dataRowInterval = dataRange.rowInterval(),
                    // the new visible row intervals
                    visibleIntervals = rowIntervals.clone(true).move(dataRowInterval.first),
                    // the new hidden row intervals
                    hiddenIntervals = new IntervalArray(dataRowInterval).difference(visibleIntervals),
                    // all intervals in a single array
                    intervals = null;

                // add visibility property to the modified interval arrays, before merging and sorting the arrays
                visibleIntervals.forEach(function (interval) { interval.attrs = { visible: true, filtered: false }; });
                hiddenIntervals.forEach(function (interval) { interval.attrs = { visible: false, filtered: true }; });

                // create a single sorted array of row intervals
                intervals = visibleIntervals.concat(hiddenIntervals).sort();

                // create operations and update all affected rows
                return sheetModel.generateChangeIntervalOperations(generator, intervals, false);
            });
        }

        // protected methods --------------------------------------------------

        /**
         * Returns the internal contents of this collection, needed for cloning
         * into another collection.
         *
         * @internal
         *  Used during clone construction. DO NOT CALL from external code!
         *
         * @returns {Object}
         *  The internal contents of this collection.
         */
        this.getCloneData = function () {
            return { tableModels: tableModels };
        };

        /**
         * Clones all contents from the passed collection into this collection.
         *
         * @internal
         *  Used by the class SheetModel during clone construction. DO NOT CALL
         *  from external code!
         *
         * @param {TableCollection} collection
         *  The source collection whose contents will be cloned into this
         *  collection.
         *
         * @returns {TableCollection}
         *  A reference to this instance.
         */
        this.cloneFrom = function (collection) {

            var // the internal contents of the source collection
                cloneData = collection.getCloneData();

            // clone the contents of the source collection
            _.each(cloneData.tableModels, function (tableModel) {
                insertTableModel(tableModel.clone(sheetModel));
            });

            return this;
        };

        /**
         * Callback handler for the document operation 'insertTable'. Creates
         * and inserts a new table into this collection.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'insertTable' document operation.
         */
        this.applyInsertOperation = function (context) {

            // get table name and map key from operation (table must not exist, throws on error)
            var modelData = getModelData(context, { insert: true });

            // check that the cell range does not cover another table
            var tableRange = context.getRange();
            context.ensure(!rangeCoversAnyTable(tableRange), 'range covers another table');

            // create and insert the new table model, notify listeners
            modelData.model = new TableModel(sheetModel, modelData.name, tableRange, context.getOptObj('attrs'));
            insertTableModel(modelData.model);
        };

        /**
         * Callback handler for the document operation 'deleteTable'. Deletes
         * an existing table from this collection.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'deleteTable' document operation.
         */
        this.applyDeleteOperation = function (context) {

            // resolve the table model addressed by the operation (throws on error)
            var modelData = getModelData(context);

            // delete the table model and notify listeners
            deleteTableModel(modelData.key);
        };

        /**
         * Callback handler for the document operation 'changeTable'. Changes
         * the position and/or attributes of an existing table.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'changeTable' document operation.
         */
        this.applyChangeOperation = function (context) {

            // resolve the table model addressed by the operation (throws on error)
            var modelData = getModelData(context);

            // change table range
            var tableRange = context.getOptRange();
            if (tableRange) {
                // check that the cell range does not cover another table
                context.ensure(!rangeCoversAnyTable(tableRange, modelData.model), 'range covers another table');
                // set the new ranges at the table model
                modelData.model.setRange(tableRange);
            }

            // change all other table attributes (change events will be forwarded)
            if (context.has('attrs')) {
                modelData.model.setAttributes(context.getObj('attrs'));
            }
        };

        /**
         * Callback handler for the document operation 'changeTableColumn'.
         * Changes the attributes of a column in an existing table.
         *
         * @param {SheetOperationContext} context
         *  A wrapper representing the 'changeTableColumn' document operation.
         */
        this.applyChangeColumnOperation = function (context) {
            getModelData(context).model.applyChangeOperation(context);
        };

        // public methods -----------------------------------------------------

        /**
         * Returns whether the specified table exists in this collection.
         *
         * @param {String} tableName
         *  The name of the table to check for. The empty string addresses the
         *  anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @returns {Boolean}
         *  Whether the specified table exists in this collection.
         */
        this.hasTable = function (tableName) {
            return tableName.toUpperCase() in tableModels;
        };

        /**
         * Returns the model of the table with the specified name.
         *
         * @param {String} tableName
         *  The name of the table. The empty string addresses the anonymous
         *  table range used to store filter settings for the standard auto
         *  filter of the sheet.
         *
         * @returns {TableModel|Null}
         *  The model of the table with the specified name; or null, if no
         *  table exists with that name.
         */
        this.getTable = function (tableName) {
            return tableModels[tableName.toUpperCase()] || null;
        };

        /**
         * Invokes the passed iterator function for all table models contained
         * in this collection.
         *
         * @param {Function} callback
         *  The callback function invoked for all table models. Receives the
         *  current table model instance as first parameter. If the callback
         *  returns the Utils.BREAK object, the iteration process will be
         *  stopped immediately.
         *
         * @returns {Utils.BREAK|Undefined}
         *  A reference to the Utils.BREAK object, if the callback has returned
         *  Utils.BREAK to stop the iteration process, otherwise undefined.
         */
        this.iterateTables = function (callback) {
            return _.any(tableModels, function (tableModel) {
                return callback.call(self, tableModel) === Utils.BREAK;
            }) ? Utils.BREAK : undefined;
        };

        /**
         * Returns all table models in this collection as plain array.
         *
         * @returns {Array<TableModel>}
         *  All table models in this collection as plain array.
         */
        this.getAllTables = function () {
            return _.values(tableModels);
        };

        /**
         * Returns the model of the table covering the specified cell.
         *
         * @param {Address} address
         *  The address of a cell.
         *
         * @returns {TableModel|Null}
         *  The model of the table at the specified position; or null, if the
         *  cell is not part of a table.
         */
        this.findTable = function (address) {
            return _.find(tableModels, function (tableModel) {
                return tableModel.containsCell(address);
            }) || null;
        };

        /**
         * Returns the models of all tables overlapping with the specified cell
         * range.
         *
         * @param {Range} range
         *  The address of a cell range.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.active=false]
         *      If set to true, the result returned by this method will be
         *      restricted to tables with activated filters/sorting (see method
         *      TableModel.isFilterActive() for details).
         *
         * @returns {Array}
         *  The models of all tables overlapping with the specified cell range.
         */
        this.findTables = function (range, options) {

            // whether to only find tables with active filter/sorting
            var active = Utils.getBooleanOption(options, 'active', false);

            return _.filter(tableModels, function (tableModel) {
                return tableModel.overlapsRange(range) && (!active || tableModel.isFilterActive());
            });
        };

        // operation generators -----------------------------------------------

        /**
         * Generates and applies a single 'insertTable' operation to insert a
         * new table range into this collection.
         *
         * @param {String} tableName
         *  The name for the new table to be inserted. The empty string refers
         *  to the anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @param {Range} range
         *  The address of the cell range covered by the new table.
         *
         * @param {Object} [attributes]
         *  The initial attribute set for the table.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the generated operations have
         *  been applied successfully, or that will be rejected with an object
         *  with 'cause' property set to one of the following error codes:
         *  - 'table:duplicate': The sheet already contains a table with the
         *      passed name.
         *  - 'table:overlap': The passed cell range covers an existing table.
         *  - 'operation': Internal error while applying the operation.
         */
        this.insertTable = function (tableName, range, attributes) {

            // check existence of the table range (TODO: table names must be unique in entire document)
            if (this.hasTable(tableName)) { return SheetUtils.makeRejected('table:duplicate'); }

            // check that the range does not overlap with an existing table
            if (this.findTables(range).length > 0) { return SheetUtils.makeRejected('table:overlap'); }

            // generate an 'insertTable' operation, and more dependent operations
            return sheetModel.createAndApplyOperations(function (generator) {

                // 'real' tables must not contain merged cells (neither in header, footer, nor data rows)
                if (tableName.length > 0) {
                    generator.generateRangeOperation(Operations.MERGE_CELLS, range, { type: 'unmerge' });
                }

                // create the 'insertTable' operation
                var properties = range.toJSON();
                if (_.isObject(attributes)) { properties.attrs = attributes; }
                generator.generateTableOperation(Operations.INSERT_TABLE, tableName, properties);

                // bug 36152: auto filter will hide drop-down buttons in header cells covered by
                // merged ranges (but only if the cells are merged before creating the auto filter)
                // - the drop-down button in the LAST column of a merged range remains visible
                if (tableName.length === 0) {
                    sheetModel.getMergeCollection().getMergedRanges(range.header()).forEach(function (mergedRange) {
                        var firstCol = Math.max(mergedRange.start[0], range.start[0]),
                            lastCol = Math.min(mergedRange.end[0], range.end[0]) - 1;
                        for (var col = firstCol; col <= lastCol; col += 1) {
                            generator.generateTableOperation(Operations.CHANGE_TABLE_COLUMN, tableName, { col: col - range.start[0], attrs: { filter: { hideMergedButton: true } } });
                        }
                    });
                }
            });
        };

        /**
         * Refreshes the specified table according to its current filter
         * settings, by generating operations for showing/hiding the affected
         * rows.
         *
         * @param {String} tableName
         *  The name of the table to be refreshed. The empty string refers to
         *  the anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the generated operations have
         *  been applied successfully, or that will be rejected with an object
         *  with 'cause' property set to one of the following error codes:
         *  - 'table:missing': The sheet does not contain a table with the
         *      passed name.
         *  - 'rows:overflow': The current filter settings of the table would
         *      modify (show or hide) too many rows at once.
         *  - 'operation': Internal error while applying the operation.
         */
        this.refreshTable = function (tableName) {
            // send operations without applying them (done manually during creation)
            return createAndSendTableOperations(tableName, generateRefreshTableOperations);
        };

        /**
         * Generates and applies a single 'changeTable' operation to modify the
         * specified table range in this collection.
         *
         * @param {String} tableName
         *  The name of the table to be modified. The empty string refers to
         *  the anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @param {Object} attributes
         *  The new attributes to be set for the table.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the generated operations have
         *  been applied successfully, or that will be rejected with an object
         *  with 'cause' property set to one of the following error codes:
         *  - 'table:missing': The sheet does not contain a table with the
         *      passed name.
         *  - 'operation': Internal error while applying the operation.
         */
        this.changeTable = function (tableName, attributes) {

            // send operations without applying them (done manually during creation)
            return createAndSendTableOperations(tableName, function (generator, tableModel) {

                // generate a single 'changeTable' operation, and immediately modify the table
                generator.generateTableOperation(Operations.CHANGE_TABLE, tableName, { attrs: attributes });
                docModel.invokeOperationHandler(generator.getLastOperation());

                // refresh the table with current filter settings
                return generateRefreshTableOperations(generator, tableModel);
            });
        };

        /**
         * Generates and applies a single 'deleteTable' operation to delete the
         * specified table range from this collection. Additionally, if the
         * table contains active filter rules, all operations will be generated
         * that make the affected hidden rows visible again.
         *
         * @param {String} tableName
         *  The name of the table to be deleted. The empty string refers to the
         *  anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the generated operations have
         *  been applied successfully, or that will be rejected with an object
         *  with 'cause' property set to one of the following error codes:
         *  - 'table:missing': The sheet does not contain a table with the
         *      passed name.
         *  - 'operation': Internal error while applying the operation.
         */
        this.deleteTable = function (tableName) {

            // send operations without applying them (done manually during creation)
            return createAndSendTableOperations(tableName, function (generator, tableModel) {

                // silently deactivate the filters in the table, and refresh the
                // table (this makes all filtered rows visible)
                tableModel.setAttributes({ table: { filtered: false } }, { notify: 'never' });
                var promise = generateRefreshTableOperations(generator, tableModel);

                // ignore 'rows:overflow' error (leave rows hidden, but delete the table anyways)
                promise = promise.then(null, function (result) {
                    return (Utils.getStringOption(result, 'cause') === 'rows:overflow') ? $.when() : result;
                });

                // add the 'deleteTable' operation, and immediately delete the table
                promise.done(function () {
                    generator.generateTableOperation(Operations.DELETE_TABLE, tableName);
                    docModel.invokeOperationHandler(generator.getLastOperation());
                });

                return promise;
            });
        };

        /**
         * Generates and applies a 'changeTableColumn' operation with the
         * passed filtering and sorting attributes, and executes the new filter
         * settings by generating operations for showing/hiding the affected
         * rows.
         *
         * @param {String} tableName
         *  The name of the table to be modified. The empty string refers to
         *  the anonymous table range used to store filter settings for the
         *  standard auto filter of the sheet.
         *
         * @param {Number} tableCol
         *  The zero-based index of the table column to be modified, relative
         *  to the cell range covered by the table.
         *
         * @param {Object} attributes
         *  The attributes to be set for the specified table column.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved when the generated operations have
         *  been applied successfully, or that will be rejected with an object
         *  with 'cause' property set to one of the following error codes:
         *  - 'table:missing': The sheet does not contain a table with the
         *      passed name.
         *  - 'rows:overflow': The passed filter settings would modify (show or
         *      hide) too many rows at once.
         *  - 'operation': Internal error while applying the operation.
         */
        this.changeTableColumn = function (tableName, tableCol, attributes) {

            // send operations without applying them (done manually during creation)
            return createAndSendTableOperations(tableName, function (generator, tableModel) {

                // store the new table column attributes
                generator.generateTableOperation(Operations.CHANGE_TABLE_COLUMN, tableName, { col: tableCol, attrs: attributes });
                docModel.invokeOperationHandler(generator.getLastOperation());

                // always refresh the filter (also if no attributes have been changed
                // actually, the cell contents may have changed in the meantime)
                return generateRefreshTableOperations(generator, tableModel);
            });
        };

        // initialization -----------------------------------------------------

        // update ranges after inserting/deleting columns or rows in the own sheet
        sheetModel.registerTransformationHandler(transformRanges);

        // destroy all class members
        this.registerDestructor(function () {
            _.invoke(tableModels, 'destroy');
            self = docModel = sheetModel = tableModels = null;
        });

    } // class TableCollection

    // exports ================================================================

    // derive this class from class ModelObject
    return ModelObject.extend({ constructor: TableCollection });

});
