/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Michael Nimz <michael.nimz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/view/popup/namedrangeslayermenu', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/control/label',
    'io.ox/office/tk/control/button',
    'io.ox/office/baseframework/view/popup/layermenu',
    'io.ox/office/baseframework/view/toolbar',
    'io.ox/office/spreadsheet/view/labels',
    'io.ox/office/spreadsheet/view/control/namedrangeitem',
    'gettext!io.ox/office/spreadsheet/main'
], function (Utils, Label, Button, LayerMenu, ToolBar, Labels, NamedRangeItem, gt) {

    'use strict';

    // class NamedRangesLayerMenu =============================================

    /**
     * A layer menu with all named ranges.
     *
     * @constructor
     *
     * @extends LayerMenu
     */
    function NamedRangesLayerMenu(docView) {

        var // self reference
            self = this,

            // the document model
            docModel = docView.getDocModel(),

            // the collection of all defined names
            nameCollection = docModel.getNameCollection();

        // base constructor ---------------------------------------------------

        LayerMenu.call(this, docView, /*#. dialog title */ gt.pgettext('named ranges', 'Named Ranges'), {
            classes: 'defined-names',
            rootContainerNode: docView.getApp().getWindowNode()
        });

        // private methods ----------------------------------------------------

        /**
         * Refreshes the pop-up content. First, the component will be released.
         * After that, a new component and all groups will be added again.
         */
        function refreshContent() {

            // the models of all visible defined names in the document (TODO: sheet names)
            var nameModels = nameCollection.getNameModels({ skipHidden: true, sort: true });

            // remove all old contents form the pop-up menu
            self.destroyAllGroups();

            // add a special entry, if no names are available at all
            if (nameModels.length === 0) {
                self.addGroup(null, new Label({ label: gt.pgettext('named ranges', 'No named ranges defined.') }));
                self.addSeparator();
            } else {
                nameModels.forEach(function addListEntry(nameModel) {
                    self.addGroup(null, new NamedRangeItem(docView, nameModel));
                    self.addSeparator();
                });
            }

            // add an extra button to define new named ranges
            self.addGroup('name/insert/dialog', new Button({ icon: 'fa-plus', label: gt.pgettext('named ranges', 'Add new named range') }));
        }

        /**
         * Moves the menu to a default position when it has been shown for the
         * first time.
         */
        function setDefaultPosition() {
            var pos = Utils.getNodePositionInPage(docView.getAppPaneNode());
            self.setNodePosition(pos.left + (pos.width - self.getNode().outerWidth()) - 40, pos.top - 20);
        }

        // initialization -----------------------------------------------------

        // add event handlers for the menu
        this.on('popup:beforeshow', refreshContent);

        // delete all menu entries when hiding the menu (entries are event listeners)
        this.one('popup:hide', this.destroyAllGroups.bind(this));

        // listen to all changes happens in the name collection
        this.listenTo(nameCollection, 'triggered', refreshContent);

        // set default position on first display of the menu
        this.one('popup:show', setDefaultPosition);

        // hide the menu automatically in read-only mode
        this.listenTo(docModel, 'change:editmode', function (event, editMode) {
            if (!editMode) { self.hide(); }
        });

        // destroy all class members
        this.registerDestructor(function () {
            self = docModel = nameCollection = docView = null;
        });

    } // class NamedRangesLayerMenu

    // exports ================================================================

    // derive this class from class LayerMenu
    return LayerMenu.extend({ constructor: NamedRangesLayerMenu });

});
